// ModelAPI.h

#ifndef MODEL_API_H
#define MODEL_API_H

#ifndef BYTE
#define BYTE unsigned char 
#endif


// levels of ode solver from fastest to simplest
#define ODELEVEL_MATEXP		6	// best: matrix exponent with constant jacobian
#define ODELEVEL_SODAJAC	5	// LSODA with user-supplied jacobian
#define ODELEVEL_SODA		4	// LSODA with finite-difference jacobian
#define ODELEVEL_NONSTIFF	3	// non-stiff solver without jacobian
#define ODELEVEL_STIFFJAC	2	// stiff solver with user-supplied jacobian
#define ODELEVEL_STIFF		1	// stiff solver with finite-difference jacobian

#define ODESTATE_START		1	// value of ISTATE when subject starts
#define ODESTATE_UNDERWAY	2	// returned value of ISTATE after a successful step
#define ODESTATE_CHANGE		3	// set ISTATE to this after dose or covariate change (but only in LSODA)

typedef enum tag_OpCode {
	OP_NUL = 0,
	OP_ID1,
	OP_ID2,
	OP_ID3,
	OP_ID4,
	OP_ID5,
	OP_NUM,
	OP_DT,
	OP_COVR,
	OP_DCOVR,	// backward delta covr
	OP_FCOVR,	// forward delta covr
	OP_BOLUS1,
	OP_BOLUS2,
	OP_INF1,
	OP_INF2,
	OP_RESET,
	OP_STRTOBS,
	OP_OBS,
	OP_OBS_BQL,
	OP_ENDOBS,
	OP_RP,
	OP_QUEST,
	OP_LP,
	OP_CLR,		// clear the stack
	OP_CLR_RESET,
	OP_INIT1,	// initialize subject after covariates initialized
	OP_ENDROW,	// mark end of input row. needed for tables
	OP_EVENT,
	OP_RPT,
	OP_SS,
	OP_SS1,
	OP_SS2,
	OP_SSEND,
	OP_SSNM,
	OP_ADDL,
	OP_ADDLNM,
	OP_DEF,
	OP_COMMA,
	OP_ADD,
	OP_SUB,
	OP_MUL,
	OP_DIV,
	OP_POW,
	OP_DUP,
	OP_SET,
	OP_REF,
} OpCode;

#ifdef __cplusplus
extern "C" {
#endif

typedef double ( fGetUniform) (void);
typedef double ( fGetNormal) (void);
typedef void   ( fOnObserve)  (double _t, double _tLastDose, const char* nm, int iCat, double dPred, double dPredObs, double dRawObs, double dSig, double dSqrVF, int bBql);
typedef void   ( fOnId)       (double dId1, double dId2, double dId3, double dId4, double dId5);
typedef void   ( fAfterLastEvent) (void);
typedef void   ( fOnEndId)    (void);
typedef void   ( fOnCovariate)(double _t, double _tLastDose, const char* nm, double dVal);
typedef void   ( fOnInit1)    (void);
typedef void   ( fOnEndRow)   (void);
typedef void   ( fOnStripDose)(double _t, double _tLastDose, const char* nm, double dVal);
typedef void   ( fOnBolus)    (double _t, double _tLastDose, int iPath, const char* nm, double dAmt);
typedef void   ( fOnInfusion) (double _t, double _tLastDose, int iPath, const char* nm, double dAmt, double dRate);
typedef void   ( fOnSS)    (void);
typedef void   ( fOnSSi)    (int iInSS);
//typedef void   ( fOnDT)       (double dt);
//typedef void   ( fOnSetTime)  (double _t);
typedef void   ( fOnEvent)    (double _t, double _tLastDose, const char* nm, double censor);
typedef void   ( fOnReset)    (void);
typedef void   ( fOnTime)     (double _t, double _tLastDose);

typedef int   ( fIsEnabled)  (int _iBitNumber);

typedef struct tag_CallbackTable {
	fGetUniform		* _pfGetUniform;
	fGetNormal		* _pfGetNormal;
	fOnObserve		* _pfOnObserve;
	fOnId			* _pfOnId;
	fAfterLastEvent	* _pfAfterLastEvent;
	fOnEndId		* _pfOnEndId;
	fOnCovariate	* _pfOnCovariate;
	fOnStripDose	* _pfOnStripDose;
	fOnInit1		* _pfOnInit1;
	fOnEndRow		* _pfOnEndRow;
	fOnBolus		* _pfOnBolus;
	fOnInfusion		* _pfOnInfusion;
	fOnSS			* _pfOnSS;
	fOnSSi			* _pfOnSSi;
	fOnEvent		* _pfOnEvent;
	fOnReset		* _pfOnReset;
	fOnTime			* _pfOnTime;
} CallbackTable;

typedef void   ( fODESolve)(int iODELevel
	, int* piState
	, int neq
	, double* ptOuter
	, double dt
	, double y[]
	, void (* calcderiv)(int*, double*, double*, double*)
	, void (* addIncrements)(double*, double, double)
	, int bTimeAdvances
	, void (* jacobian)(int*, double*, double*, int*, int*, double*, int*)
	);
typedef struct tag_ODECallbackTable {
	fODESolve		* _pfODESolve;
	fIsEnabled      * _pfIsEnabled;
} ODECallbackTable;

// delay modes for bolus doses
#define DBM_ADD_INCREMENTS 1
#define DBM_SHORT_INFUSION 2
#define DBM_DOSE_TABLE  3

// pointer to routine to that can be called for diagnostic purposes
typedef void   ( fDebug)    (int i, double d);

#ifdef __cplusplus
}; // extern "C"
#endif

// plot variable flags
#define _SWEEP_AVAIL (1<<0)
#define _POINT_AVAIL (1<<1)

#define _CAN_LOG     (1<<2)
#define _CAN_TRELLIS (1<<3)
#define _CAN_OVERLAY (1<<4)

#define _TYP_CONTIN  (1<<8)
#define _TYP_MULTI   (2<<8)
#define _TYP_EVENT   (3<<8)
#define _TYP_COUNT   (4<<8)

#ifdef __cplusplus
extern "C" {
#endif

double lnorm(double x, double stdev);

// Methods defined in MUtil.c
void _setOnTrajectory(int b);
void  SetDbgCallback(fDebug* _p);
void  SetODELevel(int* _piODELevel);
void  ModelBreakPoint();
const char*  GetErrorMsg();
int  GetNFixef();
int  GetNSecondary();
int  GetNError();
int  GetNRanef1();
int  GetNRanef2();
int  GetNRanef3();
int  GetNRanef4();
int  GetNRanef5();
int  GetInSS();
int  MultiLevelRanef();
int  AllowGaussianFit();
int  AllowLogTransform();
void  SetEnableLogTransform(int* _pbEnableLogTransform);
int  TrulyTimeBased();
int  ModelUniqueId();
void  SetBreakTime(double* pdTime);
void  SetLogFloor(double* _lf);
void  SetPredArrays(	
	int* __nPrediction
	, double* __daXPrediction
	, double* __taLastDose
	, int  * __iaWhichDose
	, int  * __iaWhichReset
	, double* __daYPrediction
	, double* __daYObservation
	, double* __daVarianceFunction
	, int  * __iaWhichObs
	);
void  InitHist();
void  EnableHistBreak(int *pb);
void  RecordHist(double d);
void  ClearActions();
void  SetMaxTime(double d);
double  GetMaxTime();
void  GetFixefName(int *pi, char* nm);
void  SetArgsOK(double** _args, int *_pbOK, int* _piState);
void  LLOneSubject(
	  int bGenGradients
	, int* pnObs
	, const unsigned char* *pb
	, double* pdId1
	, double* pdId2
	, double* pdId3
	, double* pdId4
	, double* pdId5
	, double** args
	, int* pbOK
	, double dLL[]
	, double dGrad[]
	);

void  PredOneSubject(
	  int* pnObs
	, const BYTE* *pb
	, double* pdId1, double* pdId2, double* pdId3, double* pdId4, double* pdId5
	, double** args
	, int* pbOK
	, int* __nPrediction
	, double* __daXPrediction, double* __taLastDose, int* __iaWhichDose, int* __iaWhichReset, double* __daYPrediction, double* __daYObservation, double* __daVarianceFunction
	, int* __iaWhichObs
	);
void  SimOneSubject(
	const BYTE* *pb
	, double* pdId1
	, double* pdId2
	, double* pdId3
	, double* pdId4
	, double* pdId5
	, double** args
	, int* pbOK
	, int bSampleEpsilon
	, int bForSimTbl
	, int bForVPC
	);
void  Init1OneSubject(
	const BYTE* *pb
	, double* pdId1
	, double* pdId2
	, double* pdId3
	, double* pdId4
	, double* pdId5
	, double** args
	, int* pbOK
	);
void  ScanOneSubject(
	const BYTE* *pb
	);
void ScanOneSubjectCountObs(
	const BYTE* *pb
	, int* pnObs
	, double* pdMaxTime
	);
void ScanOneSubjectGetCovariate(
	const BYTE* *pb
	, const char* sCovName
	, double* pVal
	);

// Methods defined in Model.c
void GetFixefName(int *_pi, char* _nm);
void GetFixefUnits(int *_pi, char* _nm);
void GetSecondaryName(int *_pi, char* _nm);
void GetSecondary(int* n, double* sec);
void InitEnable();	// Initialize Enablement
void SetUsingGaussianAlgorithm(int* _pBool);
int GetNThetaOld();	// get number of theta minus possibly 1 sigma to estimate
int GetNTheta();	// get number of theta to estimate
int GetNSigma();	// get number of sigma to estimate, 0 or 1
void GetRanef1Name(int *_pi, char* _nm);
void GetRanef2Name(int *_pi, char* _nm);
void GetRanef3Name(int *_pi, char* _nm);
void GetRanef4Name(int *_pi, char* _nm);
void GetRanef5Name(int *_pi, char* _nm);
void GetErrorName(int *_pi, char* _nm);
void SetArgsOK(double** __args, int *__pbOK, int *__piState);
void SetODECallbacks(ODECallbackTable* _p);
void SetCallbacks(CallbackTable* _p);
void ClearUnits();
void SetVarUnits(const char* _nm, const char* _units);
void GetPlotVarInfo(int *_pi, int* _pflags, char* _nmObs, char* _nmPred);
int GetNPlotVar();
int GetNumCovariates();
int GetCovariateNumber(const char* _nm);
void GetCovariateName(int* _pi, char* _nm);
void GetCovariateCategorical(int* _pi, int* _cat);
void GetCovariateDirection(int* _pi, int* _dir);
void SetCovariateInterp(double _t, const char* _nm, double _v1, double _dt, double _v2, int _bForward);
int GetStripDoseName(char* _nm);
int GetNumStructural();
int GetStructuralNumber(const char* _nm);
void GetStructuralName(int* _pi, char* _nm);
void GetVarUnits(const char* _nm, char* _units);
void GetNumCVMean(int* _pn);
void GetCVMeanName(int* _pi, char* _nm);
void GetCVMean(int* _pi, double* _pv);
void SetCVMean(int* _pi, double* _pv);
void GetNumCVMedian(int* _pn);
void GetCVMedianName(int* _pi, char* _nm);
void GetCVMedian(int* _pi, double* _pv);
void SetCVMedian(int* _pi, double* _pv);
void RestoreFixefLowInitHighFromModel();
void SetFixefLowInitHigh(const char* _nm, double* _fixefLow, double* _fixefInit, double* _fixefHigh);
void GetFixefLowInitHigh(const char* _nm, double* _fixefLow, double* _fixefInit, double* _fixefHigh);
void GetThetaLimits(double* _fixefLow, double* _fixefHigh);
void InitFixefErrorEstimate();
void GetFixefErrorInitial(double* _fixef);
void GetFixefErrorLimits(double* _low, double* _high);
void GetFixefValue(int* _iFixef, double* _dValue);
void GetInitialStdev(double* _svec);
void GetInitialOmega5(double* _omat);
void TruncateOmega5(double* _omat);
void UnpackParameters(double* _thvec, double* _omat, double* _pack, int* _ldOmat, int* _nPack);
void UnpackParameters1(double* _thvec, double* _omat, double* _svec, double* _pack, int* _ldOmat, int* _nPack);
void   PackParameters(double* _thvec, double* _omat, double* _pack, int* _ldOmat, int* _nPack);
void   PackParameters1(double* _thvec, double* _omat, double* _svec, double* _pack, int* _ldOmat, int* _nPack);
void   TypXParameters(double* _thvec, double* _omat, double* _pack, int* _ldOmat, int* _nPack);
void   TypXParameters1(double* _thvec, double* _omat, double* _svec, double* _pack, int* _ldOmat, int* _nPack);
void UnpackOmega(double* _omat, double* _pack, int* _ldOmat, int* _nPack);
void   PackOmega(double* _omat, double* _pack, int* _ldOmat, int* _nPack);
void   TypXOmega(double* _omat, double* _pack, int* _ldOmat, int* _nPack);
void GetFixefErrToThetaMap(int* _iaTheta, int* _nth);
void GetRanefToEtaMap(int* _iaEta, int* _nth);
void UnpackTheta(double* _fixef, double* _thvec);
void   PackTheta(double* _fixef, double* _thvec, int* _nth);
void   TypXTheta(double* _fixef, double* _thvec, int* _nth);
void   GetNPack(int* _nPack);
void   GetNPackOmega(int* _nPack);
void   GetNSeqOmega(int* _nSeq);
void   GetOmegaSequenceInfo(int* _iSeq1, int* _nEta, int* _nSame);
void   GetNPackRanCov(int* _nPack);
void   GetPackedIndexRanCov(int* _iRan, int* _jRan, int* _index);
void SetMainSigma(double* _sig);
double GetObsSigma(int* _piWhichObs);
void GetJacobianInfo(int* _bUseAnagrad, int* _piJacobianInfo);
void GetMaxODELevel(int* _bUseAnagrad, int* _piMaxODELevel);
int GetNumCompartments();
int GetCompartmentNumber(const char* _nm);
void GetCompartmentName(int* _pi, char* _nm);
int GetVarValue(const char* _nm, double* _pv);
void SetVarValue(const char* _nm, double* _pv);
void GetSaemCMatrixSize(int* nBaseEta, int* nExtraEta, int* nTheta);
void GetSaemCMatrix(double* daCMat);
int GetCovariateIsInCMatrix(const char* name);
void GetSaemThetaInfo(int* iaCColToITheta, int* baCColIsLog);
void GetSaemNaivePoolThetas(int* nNPTheta, int* iaTheta);
int CanUseSAEM();
void DoEndObsActions();
int GetNIntegGradientDeriv();
void GetLLGradientEta(
	  int* _pnObs
	, const unsigned char* *ppb
	, double* _pdId1, double* _pdId2, double* _pdId3, double* _pdId4, double* _pdId5
	, double _dLL[]
	, double _dGrad[]
	);
void NewPredOneSubject(
	  int* _pnObs
	, const unsigned char* *ppb
	, double* _pdId1, double* _pdId2, double* _pdId3, double* _pdId4, double* _pdId5
	, double** _args
	, int* _pbOK
	, int* _nPrediction
	, double* __daXPrediction
	, double* __taLastDose
	, int  * __iaWhichDose
	, int  * __iaWhichReset
	, double* __daYPrediction
	, double* __daYObservation
	, double* __daVarianceFunction
	, double* _dYdK // comes out as _daYdK(iEta, iObs)
	, double* _dSdK // comes out as _daSdK(iEta, iObs)
	, int* _iaWhichObs
	);
int GetNumObservations();
int GetObservationNumber(const char* _nm);
void GetObservationName(int* _pi, char* _nm);
int GetObservationType(int* _pi);
void GetObservationBqlLimit(int* _pi, double* _bqllimit);
int GetPredObsError(int _iObs);
double GetPredVFCorrectionFactor(int _iError);
void GetPredName(int* _iWhichObs, char* _nm);
void GetObsName(int* _iWhichObs, char* _nm);
void GetIVarName(int* _iWhichObs, char* _nm);
void GetObsVF(int* _iWhichObs, double* _vf);
void GetNumGaussianObs(int* _nGaussianObs);
void GetNumCategories(const char* _nm, int* pNCat);
void SimulateForSimTbl(double _t, const char* _nm, double* _pObs, double* _pPred);

#ifdef __cplusplus
}; // extern "C"
#endif

#endif	// MODEL_API_H

